import boto3
import json
import os
from contextlib import closing
import pickle

VIS2IPA = {"p": "BILABIAL",
            "f": "LABIODENTAL",
            "T": "INTERDENTAL",
            "s": "DENTAL_ALVEOLAR",
            "t": "DENTAL_ALVEOLAR",
            "S": "POSTALVEOLAR",
            "r": "POSTALVEOLAR",
            "k": "VELAR_GLOTTAL",
            "i": "CLOSE_FRONT_VOWEL",
            "u": "CLOSE_BACK_VOWEL",
            "@": "MID_CENTRAL_VOWEL",
            "a": "OPEN_FRONT_VOWEL",
            "e": "OPEN_FRONT_VOWEL",
            "E": "OPEN_FRONT_VOWEL",
            "o": "OPEN_BACK_VOWEL",
            "O": "OPEN_BACK_VOWEL",
            "sil": "IDLE"}


class PollyTTS:

    def __init__(self):
        self.tts = boto3.client('polly')

    def gen_audio_and_visemes(self, text, voice_id=None, fname=None):
        '''
        generates an audio file that says the text of _text_ in the voice of _voice_id_

        :param: text - the text that the robot should speak
        :param: voice_id - the voice that the robot should speak in
        :param: fname - the name of the file that the audio should be saved to
        '''
        if voice_id is None:
            voice_id = 'Justin'
            
        if fname is None:   
            fname = f"pylips_phrases/{voice_id}_output.mp3"
        else:
            #if it was already cached, return it, otherwise, generate it and return it
            fname = f"pylips_phrases/{fname}.mp3"
            if os.path.exists(fname):
                times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
                return fname, times, visemes

        # Synthesize speech
        response = self.tts.synthesize_speech(
            TextType='ssml',
            Text=f"<speak>{text}</speak>",
            OutputFormat='mp3',
            VoiceId=voice_id
        )

        audio_stream = response["AudioStream"]

        if not os.path.isdir('pylips_phrases'):
            os.mkdir('pylips_phrases')
             
        with open(fname, "wb") as file:
            file.write(audio_stream.read())

        # Synthesize visemes
        # Synthesize speech with viseme output
        response = self.tts.synthesize_speech(
            TextType='ssml',
            Text=f"<speak>{text}</speak>",
            OutputFormat='json',
            SpeechMarkTypes=['viseme'],
            VoiceId=voice_id
        )
        
         # Process the response
        if 'AudioStream' in response:
            # Do something with the audio stream if needed
            with closing(response["AudioStream"]) as stream:
                        data = stream.read().decode('utf-8')
                        xSheet = data.split('\n')
                        xSheet = [json.loads(line) for line in xSheet if line != '']

        times = [x['time'] / 1000. for x in xSheet]
        visemes = [VIS2IPA[x['value']] for x in xSheet]

        pickle.dump((times, visemes), open(f'{fname[:-4]}.pkl', 'wb'))

        return fname, times, visemes
    
    def get_audio_and_visemes(self, fname):
        '''
        
        '''
        #if it was already cached, return it, otherwise, raise an error
        fname = f"pylips_phrases/{fname}.mp3"

        if os.path.exists(fname):
            times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
            return fname, times, visemes
        else:
            raise Exception(f'phrase {fname} does not exist')


from robot_face import RobotFace

if __name__ == '__main__':
    aimee = RobotFace(robot_name='aimee', tts_method='system', voice_id='com.apple.voice.premium.en-US.Zoe')
    aimee.tts.list_voices()

    nathan = RobotFace(robot_name='nathan', voice_id='Matthew')

    # send_message(request)
    while True:
        message = input("who will we make talk?: ")
        aimee.say(message)
            
        


import socketio
import boto3
from contextlib import closing
import json
import pickle
import pygame
import time
import os

from .polly_tts import PollyTTS
from .system_tts import SystemTTS

class RobotFace:
    def __init__(self,
                 robot_name: str = 'default',
                 server_ip: str = 'http://localhost:8000',
                 tts_method: str = 'system',
                 voice_id: str = None):
        '''
            The main class for the PyLips project.
            :param: robot_name - the identity of the robot that should be speaking
            :param: server_ip - the location of the server running the flask application.
        '''

        if os.path.exists('pylips_phrases') == False:
            os.mkdir('pylips_phrases')

        self.name = robot_name
        self.voice_id = voice_id

        self.io = socketio.Client()
        self.io.connect(server_ip)

        pygame.mixer.init()
        self.channel = 0

        tts_methods = ['polly']
        if tts_method not in tts_method:
            raise Exception(f'parameter tts_method must be one of {tts_methods}')
        
        # TODO: implement other TTS things, but for now we will use polly
        if tts_method == 'system':
            self.tts = SystemTTS()
        if tts_method == 'polly':
            self.tts = PollyTTS()

    def say(self, content):
        '''
            The main method for the RobotFace class. This method will take in a string of text and
            convert it to speech using the AWS Polly service.
            :param: content - the text that the robot should speak
        '''

        fname, times, visemes = self.tts.gen_audio_and_visemes(content, self.voice_id)

        request = { 
                'name': self.name,
                'action_type': 'say',
                'visemes': visemes,
                'times': times,
             }
        

        #load sound
        while pygame.mixer.Channel(self.channel).get_busy():
            self.channel += 1
             
        sound = pygame.mixer.Sound(fname)

        #play sound and face
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).play(sound)

    def save_file(self, content, filename):
        '''
            The main method for the RobotFace class. This method will take in a string of text and
            convert it to speech.
            :param: content - the text that the robot should speak
        '''
        self.tts.gen_audio_and_visemes(content, self.voice_id, filename)
         

    def say_file(self, filename):

        fname, times, visemes = self.tts.get_audio_and_visemes(filename)

        request = { 
                'name': self.name,
                'action_type': 'say',
                'visemes': visemes,
                'times': times,
             }
        
        #load sound
        while pygame.mixer.Channel(self.channel).get_busy():
            self.channel += 1
             
        sound = pygame.mixer.Sound(fname)

        #play sound and face
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).play(sound)


    def stop_speech(self):
        request = { 
                'name': self.name,
                'action_type': 'stop_speech',
             }
        
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).stop()
         
        
    def look(self, x, y, z, time):
        '''
        Looks to a location in the face's reference frame.
        x,y,z in mm
        time in ms
        '''
        request = {
                'name': self.name,
                'action_type': 'look',
                'location': [x,y,z],
                'time': time
            }
        self.io.emit('face_control', request)

    def release_gaze(self):
        '''
        allows robot to move eyes idly
        '''
        request = {
            'name': self.name,
            'action_type': 'release_gaze',
        }
        self.io.emit('face_control', request)

    def express(self, aus, time):
        request = {
            'name': self.name,
            'action_type': 'express',
            'aus': aus,
            'time': time
        }
        self.io.emit('face_control', request)

    def update_face(self, config):
        request = {
            'name': self.name,
            'action_type': 'update_face',
            'configuration': config
        }
        self.io.emit('face_control', request)
        time.sleep(.3)


         
